require 'redis'

require 'diplo/env'

module Diplo
  module Lock
    module_function

    def create(resource, ttl = nil)
      ttl ||= Env.fetch_int(:lock_ttl)
      token = create_lock_token
      lock = OpenStruct.new(resource: resource, token: token)

      retry_count.times do
        locked_nodes, start_time = 0, Time.now

        servers.each do |server|
          locked_nodes += 1 if lock_server(server, resource, token, ttl)
        end

        validity = validity_time(start_time, ttl)

        if locked_nodes >= quorum && validity > 0
          lock.validity = validity
          return lock
        else
          remove lock
        end

        delay
      end

      false
    end

    def remove(lock)
      servers.each do |server|
        unlock_server server, lock.resource, lock.token
      end
    end

    def servers
      @servers || begin
        config = Env.fetch(:redis)
        uris = config.gsub(' ', '').split(',')
        uris.map { |u| ::Redis.new url: u }
      end
    end

    def lock_server(redis, resource, token, ttl)
      redis.client.call [:set, resource, token, :nx, :px, ttl]
    rescue
      false
    end

    def unlock_server(redis, resource, token)
      @unlock_script ||= <<-LUA
        if redis.call("get", KEYS[1]) == ARGV[1] then
          return redis.call("del", KEYS[1])
        else
          return 0
        end
      LUA

      redis.client.call [:eval, @unlock_script, 1, resource, token]
    rescue
      # Nothing to do, unlocking is just a best-effort attempt.
    end

    def retry_count
      @retry_count || Env.fetch_int(:lock_retry_count)
    end

    def retry_count=(count)
      @retry_count = Integer(count)
    end

    def retry_delay
      @retry_delay || Env.fetch_int(:lock_retry_delay)
    end

    def retry_delay=(miliseconds)
      @retry_delay = Integer(miliseconds)
    end

    def drift_factor
      @drift_factor || Env.fetch_float(:lock_drift_factor)
    end

    def drift_factor=(factor)
      @drift_factor = Float(factor)
    end

    def lock_token_bytes
      @lock_token_bytes || Env.fetch_int(:lock_token_bytes)
    end

    def lock_token_bytes=(count)
      @lock_token_bytes = Integer(count)
    end

    def quorum
      @quorum || Env.fetch_int(:lock_quorum)
    end

    def quorum=(nodes)
      @quorum = Integer(nodes)
    end

    def create_lock_token
      SecureRandom.hex lock_token_bytes
    end

    def delay
      sleep rand(retry_delay).to_f / 1000
    end

    def validity_time(start_time, ttl)
      diff_time = Time.now - start_time
      drift = (ttl* drift_factor).to_i + 2
      ttl - (diff_time * 1000).to_i - drift
    end

  end
end